import errno
import os
import sys

from setuptools import Extension

from paver.easy import *
from paver.path import path
from paver.setuputils import setup


VERSION = (1, 0, 2, "")

info = dict(
    name="mummy",
    description="fast, efficient serialization",
    packages=["mummy", "oldmummy"],
    package_dir={'': 'python'},
    version=".".join(filter(None, map(str, VERSION))),
    author="Travis Parker",
    author_email="travis.parker@gmail.com",
    url="http://github.com/teepark/mummy",
    license="BSD",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: C",
        "Programming Language :: Python",
    ],
)

if sys.subversion[0].lower() != 'pypy':
    info['ext_modules'] = [
        Extension(
            '_mummy',
            ['python/dump.c', 'python/load.c', 'python/mummymodule.c',
                'lzf/lzf_c.c', 'lzf/lzf_d.c',
                'lib/mummy_string.c', 'lib/dump.c', 'lib/load.c'],
            include_dirs=('python', 'lzf', 'include'),
            extra_compile_args=['-Wall']),
        Extension(
            '_oldmummy',
            ['python/_old_mummy.c', 'lzf/lzf_c.c', 'lzf/lzf_d.c'],
            include_dirs=('python', 'lzf'),
            extra_compile_args=['-Wall']),
        ]

setup(**info)

MANIFEST = (
    "setup.py",
    "paver-minilib.zip",
    "include/mummy.h",
    "python/mummypy.h",
    "lzf/lzf.h",
    "lzf/lzfP.h",
)

@task
def manifest():
    path('MANIFEST.in').write_lines('include %s' % x for x in MANIFEST)

@task
@needs('generate_setup', 'minilib', 'manifest', 'setuptools.command.sdist')
def sdist():
    pass

@task
def clean():
    for p in map(path, ('mummy.egg-info', 'dist', 'build', 'MANIFEST.in')):
        if p.exists():
            if p.isdir():
                p.rmtree()
            else:
                p.remove()
    for p in path(__file__).abspath().parent.walkfiles():
        if p.endswith(".pyc") or p.endswith(".pyo"):
            try:
                p.remove()
            except OSError, exc:
                if exc.args[0] == errno.EACCES:
                    continue
                raise

@task
def test():
    sh("cd python; nosetests")

@task
def docs():
    sh("find docs/source/mummy -name *.rst | xargs touch")
    sh("cd docs; make html")



"""
mummy is the name for a data serialization format and accompanying
(de)serializer.

the format is designed to be first fast and second compact, sacrificing
human-readability in the name of those two values.

it's prefix-encoded. that means the first byte is the type, and that is
followed by the content. depending on the type, the content may begin with more
meta-data.

first byte is the type:
    0x00 null
    0x01 boolean
    0x02 char
    0x03 short
    0x04 integer
    0x05 long
    0x06 huge
    0x07 float
    0x08 shortstring
    0x18 medstring
    0x09 longstring
    0x0A short utf8 string
    0x19 medium utf8
    0x0B long utf8 string
    0x10 short list
    0x14 medium list
    0x0C long list
    0x11 short tuple
    0x15 medium tuple
    0x0D long tuple
    0x12 short set
    0x16 mediumset
    0x0E long set
    0x13 short hash
    0x17 medium hash
    0x0F long hash
    0x1A date
    0x1B time
    0x1C date & time
    0x1D time difference
    0x1E decimal
    0x1F special number

* null: no body, decodes to python None
* boolean: one byte body (0 or 1), python bool
* char, short, integer, long: 1, 2, 4, 8 byte bodies respectively, containing
        big-endian signed integers. comes out as a python int (on 32 bit
        pythons the mummy 'long' type will be a python long)
* huge: 4 byte prefix holding the length of the rest of the body, which is a
        big-endian signed integer of arbitrary size. decodes to a python long.
* float: 64-bit IEEE floating point. python float type.
* strings: 1, 2 or 4 byte prefix for (short, med, long) strings. prefix
        contains the length of the body, which is just the input bytes. decodes
        to python bytestring (python2 string, python3 bytes).
* utf8: same as strings but encodes from/decodes to python unicode strings via
        UTF-8. the length prefix is encoded length, not unicode char count.
* lists: 1, 2 or 4 byte length prefix for (short, med, long) lists. prefix in
        this case is # of contained elements, not body byte length. the body
        simply holds the mummified contents.
* tuples: same as lists, only difference is the python type
* sets: same as lists, only difference is the python type
* hashes: 1, 2 or 4 byte length prefix holds # of contained key value pairs.
        the body is mummified alternating keys and values with no separators.
        encodes from/decodes to python dict.
* date: no header, then 2 byte unsigned big-endian year, one byte month, one
        byte day. python type datetime.date
* time: no header, then one byte each for (hour, minute, second), then unsigned
        3-byte big-endian microsecond. python type is datetime.time
* date & time: after the single type byte it holds the bodies of a date and
        time mashed together. python type is datetime.datetime
* time difference: signed 4 big-endian bytes for each of days, seconds,
        microseconds. python type is datetime.timedelta.
* decimal: single byte of sign (0 or 1), 2 byte signed big-endian for decimal
        point position, 2 byte unsigned big-endian for number of digits, then
        4-bit unsigned numbers for digits alternating low 4, high 4 (always
        low 4 first). python type is decimal.Decimal.
* special numbers: single byte body, the high 4 bits value can be 1 for
        infinity or 2 for NaN, and the low 4 bits value being 1 (instead of 0)
        turns Infinity into -Infinity, or NaN into sNaN. the python class is
        decimal.Decimal, which supports all 4 of these special numbers.

the main implementation is in C, but there is a pure-python version it falls
back to if the extension is unavailable. the module-global `has_extension` is a
boolean indicating whether the C extension is in use.
"""

from __future__ import absolute_import

from .serialization import \
        loads, dumps, pure_python_loads, pure_python_dumps, has_extension
from .schemas import Message, OPTIONAL, UNION, ANY


VERSION = (1, 0, 2, "")
__version__ = ".".join(filter(None, map(str, VERSION)))


__all__ = ["loads", "dumps", "pure_python_loads", "pure_python_dumps",
        "has_extension", "Message", "OPTIONAL", "UNION", "ANY"]



import datetime
import decimal
import fractions
import itertools
import struct
import sys

try:
    import lzf
except ImportError:
    lzf = None


__all__ = ["loads", "dumps", "pure_python_loads", "pure_python_dumps",
        "has_extension"]


if sys.version_info[0] >= 3:
    from functools import reduce
    _str = unicode = str
    class string(bytes):
        def __getitem__(self, index):
            if isinstance(index, int):
                return type(self)(super().__getitem__(
                    slice(index, index + 1, None)))
            return type(self)(super().__getitem__(index))

        def __iter__(self):
            i, l = 0, len(self)
            while i < l:
                yield self[i]
                i += 1

    long = int
    map = lambda *a: list(__builtins__['map'](*a))
    iteritems = lambda d: d.items()
    iterkeys = lambda d: d.keys()
    itervalues = lambda d: d.values()
    xrange = range
    chr = lambda x: bytes([x])
    ord = lambda x: bytes.__getitem__(x, 0)
    bytify = lambda x: x.encode('ascii')
    izip = zip
else:
    iteritems = lambda d: d.iteritems()
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    bytify = lambda x: x
    bytes = string = str
    izip = itertools.izip


MAX_DEPTH = 256

MUMMY_TYPE_NULL = 0x0
MUMMY_TYPE_BOOL = 0x1
MUMMY_TYPE_CHAR = 0x2
MUMMY_TYPE_SHORT = 0x3
MUMMY_TYPE_INT = 0x4
MUMMY_TYPE_LONG = 0x5
MUMMY_TYPE_HUGE = 0x6
MUMMY_TYPE_FLOAT = 0x7
MUMMY_TYPE_SHORTSTR = 0x8
MUMMY_TYPE_LONGSTR = 0x9
MUMMY_TYPE_SHORTUTF8 = 0xA
MUMMY_TYPE_LONGUTF8 = 0xB
MUMMY_TYPE_LONGLIST = 0xC
MUMMY_TYPE_LONGTUPLE = 0xD
MUMMY_TYPE_LONGSET = 0xE
MUMMY_TYPE_LONGHASH = 0xF

MUMMY_TYPE_SHORTLIST = 0x10
MUMMY_TYPE_SHORTTUPLE = 0x11
MUMMY_TYPE_SHORTSET = 0x12
MUMMY_TYPE_SHORTHASH = 0x13

MUMMY_TYPE_MEDLIST = 0x14
MUMMY_TYPE_MEDTUPLE = 0x15
MUMMY_TYPE_MEDSET = 0x16
MUMMY_TYPE_MEDHASH = 0x17
MUMMY_TYPE_MEDSTR = 0x18
MUMMY_TYPE_MEDUTF8 = 0x19

MUMMY_TYPE_DATE = 0x1A
MUMMY_TYPE_TIME = 0x1B
MUMMY_TYPE_DATETIME = 0x1C
MUMMY_TYPE_TIMEDELTA = 0x1D

MUMMY_TYPE_DECIMAL = 0x1E
MUMMY_TYPE_SPECIALNUM = 0x1F

MUMMY_TYPE_FRACTION = 0x20

MUMMY_SPECIAL_INFINITY = 0x10
MUMMY_SPECIAL_NAN = 0x20


TYPEMAP = {
    type(None): MUMMY_TYPE_NULL,
    bool: MUMMY_TYPE_BOOL,
    float: MUMMY_TYPE_FLOAT,
}

_BIG_ENDIAN = struct.pack("!h", 1) == struct.pack("h", 1)


def _get_type_code(x):
    mapped = TYPEMAP.get(type(x))
    if mapped is not None:
        return mapped

    if type(x) is list:
        if len(x) < 256:
            return MUMMY_TYPE_SHORTLIST
        if len(x) < 65536:
            return MUMMY_TYPE_MEDLIST
        return MUMMY_TYPE_LONGLIST

    if type(x) is tuple:
        if len(x) < 256:
            return MUMMY_TYPE_SHORTTUPLE
        if len(x) < 65536:
            return MUMMY_TYPE_MEDTUPLE
        return MUMMY_TYPE_LONGTUPLE

    if type(x) in (set, frozenset):
        if len(x) < 256:
            return MUMMY_TYPE_SHORTSET
        if len(x) < 65536:
            return MUMMY_TYPE_MEDSET
        return MUMMY_TYPE_LONGSET

    if type(x) is dict:
        if len(x) < 256:
            return MUMMY_TYPE_SHORTHASH
        if len(x) < 65536:
            return MUMMY_TYPE_MEDHASH
        return MUMMY_TYPE_LONGHASH

    if type(x) is bytes:
        if len(x) < 256:
            return MUMMY_TYPE_SHORTSTR
        if len(x) < 65536:
            return MUMMY_TYPE_MEDSTR
        return MUMMY_TYPE_LONGSTR

    if type(x) is unicode:
        if len(x.encode('utf8')) < 256:
            return MUMMY_TYPE_SHORTUTF8
        if len(x.encode('utf8')) < 65536:
            return MUMMY_TYPE_MEDUTF8
        return MUMMY_TYPE_LONGUTF8

    if type(x) in (int, long):
        if -128 <= x < 128:
            return MUMMY_TYPE_CHAR
        if -32768 <= x < 32768:
            return MUMMY_TYPE_SHORT
        if -2147483648 <= x < 2147483648:
            return MUMMY_TYPE_INT
        if -9223372036854775808 <= x < 9223372036854775808:
            return MUMMY_TYPE_LONG
        return MUMMY_TYPE_HUGE

    if type(x) is datetime.date:
        return MUMMY_TYPE_DATE

    if type(x) is datetime.time:
        return MUMMY_TYPE_TIME

    if type(x) is datetime.datetime:
        return MUMMY_TYPE_DATETIME

    if type(x) is datetime.timedelta:
        return MUMMY_TYPE_TIMEDELTA

    if type(x) is decimal.Decimal:
        if x.is_nan() or x.is_infinite():
            return MUMMY_TYPE_SPECIALNUM
        return MUMMY_TYPE_DECIMAL

    if type(x) is fractions.Fraction:
        return MUMMY_TYPE_FRACTION

    raise ValueError("%r cannot be serialized" % type(x))


##
## DUMPERS
##

def _dump_none(x, depth=0, default=None):
    return bytify("")

def _dump_bool(x, depth=0, default=None):
    return bytify(x and '\x01' or '\x00')

def _dump_char(x, depth=0, default=None):
    if x < 0:
        x += 256
    return chr(x)

def _dump_uchar(x, depth=0, default=None):
    return chr(x)

def _dump_short(x, depth=0, default=None):
    return struct.pack("!h", x)

def _dump_ushort(x, depth=0, default=None):
    return struct.pack("!H", x)

def _dump_int(x, depth=0, default=None):
    return struct.pack("!i", x)

def _dump_uint(x, depth=0, default=None):
    return struct.pack("!I", x)

def _dump_long(x, depth=0, default=None):
    return struct.pack("!q", x)

def _dump_huge(x, depth=0, default=None):
    data = []
    neg = x < 0
    if neg:
        x = ~x
    while x:
        data.append(x & 0xff)
        x >>= 8
    if neg:
        data = map(lambda byte: byte ^ 0xff, data)
    if not _BIG_ENDIAN:
        data = data[::-1]
    if neg and not data[0] & 0x80:
        data = [255] + data
    if not neg and data[0] & 0x80:
        data = [0] + data
    data = map(chr, data)
    return _dump_uint(len(data)) + bytify("").join(data)

def _dump_double(x, depth=0, default=None):
    return struct.pack("!d", x)

def _dump_shortstr(x, depth=0, default=None):
    return _dump_uchar(len(x)) + x

def _dump_medstr(x, depth=0, default=None):
    return _dump_ushort(len(x)) + x

def _dump_longstr(x, depth=0, default=None):
    return _dump_uint(len(x)) + x

def _dump_shortutf8(x, depth=0, default=None):
    return _dump_shortstr(x.encode('utf8'))

def _dump_medutf8(x, depth=0, default=None):
    return _dump_medstr(x.encode('utf8'))

def _dump_longutf8(x, depth=0, default=None):
    return _dump_longstr(x.encode('utf8'))

def _dump_list(x, depth=0, default=None):
    return _dump_uint(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0)
            for item in x)

_dump_set = _dump_tuple = _dump_list

def _dump_dict(x, depth=0, default=None):
    return _dump_uint(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0) for item in
            reduce(lambda a, b: a.extend(b) or a, iteritems(x), []))

def _dump_shortlist(x, depth=0, default=None):
    return _dump_uchar(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0)
            for item in x)

_dump_shorttuple = _dump_shortset = _dump_shortlist

def _dump_shortdict(x, depth=0, default=None):
    return _dump_uchar(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0) for item in
            reduce(lambda a, b: a.extend(b) or a, iteritems(x), []))

def _dump_medlist(x, depth=0, default=None):
    return _dump_ushort(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0)
            for item in x)

_dump_medtuple = _dump_medset = _dump_medlist

def _dump_meddict(x, depth=0, default=None):
    return _dump_ushort(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0) for item in
            reduce(lambda a, b: a.extend(b) or a, iteritems(x), []))

def _dump_date(x, depth=0, default=None):
    return "".join(
            (_dump_ushort(x.year), _dump_char(x.month), _dump_char(x.day)))

def _dump_time(x, depth=0, default=None):
    if x.tzinfo is not None:
        raise ValueError("can't serialize data objects with tzinfo")
    return "".join((
        _dump_char(x.hour),
        _dump_char(x.minute),
        _dump_char(x.second),
        struct.pack("!I", x.microsecond)[-3:]))

def _dump_datetime(x, depth=0, default=None):
    return _dump_date(x.date()) + _dump_time(x.timetz())

def _dump_timedelta(x, depth=0, default=None):
    return "".join((
        _dump_int(x.days), _dump_int(x.seconds), _dump_int(x.microseconds)))

def _dump_decimal(x, depth=0, default=None):
    sign, digits, expo = x.as_tuple()
    pairs = []
    for i, dig in enumerate(digits):
        if not 0 <= dig <= 9:
            raise ValueError("invalid digit")

        if not (i & 1):
            # even
            pairs.append(0)
        else:
            # odd
            dig <<= 4
        pairs[-1] |= dig

    return (struct.pack("!bhH", sign, expo, len(digits)) +
            "".join(map(_dump_uchar, pairs)))

def _dump_specialnum(x, depth=0, default=None):
    if x.is_snan():
        return _dump_uchar(MUMMY_SPECIAL_NAN | 1)

    if x.is_nan():
        return _dump_uchar(MUMMY_SPECIAL_NAN)

    if x.is_infinite():
        return _dump_uchar(MUMMY_SPECIAL_INFINITY | int(x < 0))

def _dump_fraction(x, depth=0, default=None):
    return struct.pack("!qq", x.numerator, x.denominator)


_dumpers = {
    MUMMY_TYPE_NULL: _dump_none,
    MUMMY_TYPE_BOOL: _dump_bool,
    MUMMY_TYPE_CHAR: _dump_char,
    MUMMY_TYPE_SHORT: _dump_short,
    MUMMY_TYPE_INT: _dump_int,
    MUMMY_TYPE_LONG: _dump_long,
    MUMMY_TYPE_HUGE: _dump_huge,
    MUMMY_TYPE_FLOAT: _dump_double,
    MUMMY_TYPE_SHORTSTR: _dump_shortstr,
    MUMMY_TYPE_MEDSTR: _dump_medstr,
    MUMMY_TYPE_LONGSTR: _dump_longstr,
    MUMMY_TYPE_SHORTUTF8: _dump_shortutf8,
    MUMMY_TYPE_MEDUTF8: _dump_medutf8,
    MUMMY_TYPE_LONGUTF8: _dump_longutf8,
    MUMMY_TYPE_LONGLIST: _dump_list,
    MUMMY_TYPE_LONGTUPLE: _dump_tuple,
    MUMMY_TYPE_LONGSET: _dump_set,
    MUMMY_TYPE_LONGHASH: _dump_dict,
    MUMMY_TYPE_SHORTLIST: _dump_shortlist,
    MUMMY_TYPE_MEDLIST: _dump_medlist,
    MUMMY_TYPE_SHORTTUPLE: _dump_shorttuple,
    MUMMY_TYPE_MEDTUPLE: _dump_medtuple,
    MUMMY_TYPE_SHORTSET: _dump_shortset,
    MUMMY_TYPE_MEDSET: _dump_medset,
    MUMMY_TYPE_SHORTHASH: _dump_shortdict,
    MUMMY_TYPE_MEDHASH: _dump_meddict,
    MUMMY_TYPE_DATE: _dump_date,
    MUMMY_TYPE_TIME: _dump_time,
    MUMMY_TYPE_DATETIME: _dump_datetime,
    MUMMY_TYPE_TIMEDELTA: _dump_timedelta,
    MUMMY_TYPE_DECIMAL: _dump_decimal,
    MUMMY_TYPE_SPECIALNUM: _dump_specialnum,
    MUMMY_TYPE_FRACTION: _dump_fraction,
}

def pure_python_dumps(item, default=None, depth=0, compress=True):
    """serialize a native python object into a mummy string
    
    :param object: the python object to serialize
    :param function default:
        If the 'object' parameter is not serializable and this parameter is
        provided, this function will be used to generate a fallback value to
        serialize. It should take one argument (the original object), and
        return something serilizable.
    :param bool compress:
        whether or not to attempt to compress the serialized data (default
        True)

    :returns: the bytestring of the serialized data
    """
    if default and not hasattr(default, "__call__"):
        raise TypeError("default must be callable or None")
    if depth >= MAX_DEPTH:
        raise ValueError("max depth exceeded")
    try:
        kind = _get_type_code(item)
    except ValueError:
        if default is None:
            raise TypeError("unserializable type")
        item = default(item)
        kind = _get_type_code(item)
    data = _dumpers[kind](item, depth, default)
    datalen = len(data)
    if compress and lzf and datalen > 5:
        compressed = lzf.compress(data, datalen - 5)
        if compressed:
            data = struct.pack("!i", datalen) + compressed
            kind = kind | 0x80
    kind = _dump_char(kind)

    return kind + data


##
## LOADERS
##

def _load_none(x):
    return None, 0

def _load_bool(x):
    return bool(ord(x[0])), 1

def _load_char(x):
    num = ord(x[0])
    if num >= 128:
        num -= 256
    return num, 1

def _load_uchar(x):
    return ord(x[0]), 1

def _load_short(x):
    return struct.unpack("!h", x[:2])[0], 2

def _load_ushort(x):
    return struct.unpack("!H", x[:2])[0], 2

def _load_int(x):
    return struct.unpack("!i", x[:4])[0], 4

def _load_uint(x):
    return struct.unpack("!I", x[:4])[0], 4

def _load_long(x):
    return struct.unpack("!q", x[:8])[0], 8

def _load_huge(x):
    width = _load_uint(x)[0] + 4
    num = 0
    x = x[4:]
    neg = ord(x[0]) & 0x80
    data = map(ord, x[:width])
    if neg:
        data = map(lambda byte: byte ^ 0xff, data)
    for c in data:
        num = (num << 8) | c
    if neg:
        return -num - 1, width
    return num, width

def _load_double(x):
    return struct.unpack("!d", x[:8])[0], 8

def _load_shortstr(x):
    width = _load_uchar(x)[0] + 1
    return x[1:width], width

def _load_medstr(x):
    width = _load_ushort(x)[0] + 2
    return x[2:width], width

def _load_longstr(x):
    width = _load_uint(x)[0] + 4
    return x[4:width], width

def _load_shortutf8(x):
    str, width = _load_shortstr(x)
    return str.decode('utf8'), width

def _load_medutf8(x):
    str, width = _load_medstr(x)
    return str.decode('utf8'), width

def _load_longutf8(x):
    str, width = _load_longstr(x)
    return str.decode('utf8'), width

def _load_list(x):
    length, width = _load_uint(x)
    result = []
    for i in xrange(length):
        item, item_width = _loads(x[width:])
        result.append(item)
        width += item_width + 1
    return result, width

def _load_set(x):
    lst, width = _load_list(x)
    return set(lst), width

def _load_tuple(x):
    lst, width = _load_list(x)
    return tuple(lst), width

def _load_dict(x):
    length, width = _load_uint(x)
    result = {}
    for i in xrange(length):
        key, keywidth = _loads(x[width:])
        width += keywidth + 1
        value, valuewidth = _loads(x[width:])
        width += valuewidth + 1
        result[key] = value
    return result, width

def _load_shortlist(x):
    length, width = _load_uchar(x)
    result = []
    for i in xrange(length):
        item, item_width = _loads(x[width:])
        result.append(item)
        width += item_width + 1
    return result, width

def _load_shorttuple(x):
    lst, width = _load_shortlist(x)
    return tuple(lst), width

def _load_shortset(x):
    lst, width = _load_shortlist(x)
    return set(lst), width

def _load_shortdict(x):
    length, width = _load_uchar(x)
    result = {}
    for i in xrange(length):
        key, keywidth = _loads(x[width:])
        width += keywidth + 1
        value, valuewidth = _loads(x[width:])
        width += valuewidth + 1
        result[key] = value
    return result, width

def _load_medlist(x):
    length, width = _load_ushort(x)
    result = []
    for i in xrange(length):
        item, item_width = _loads(x[width:])
        result.append(item)
        width += item_width + 1
    return result, width

def _load_medtuple(x):
    lst, width = _load_medlist(x)
    return tuple(lst), width

def _load_medset(x):
    lst, width = _load_medlist(x)
    return set(lst), width

def _load_meddict(x):
    length, width = _load_ushort(x)
    result = {}
    for i in xrange(length):
        key, keywidth = _loads(x[width:])
        width += keywidth + 1
        value, valuewidth = _loads(x[width:])
        width += valuewidth + 1
        result[key] = value
    return result, width

def _load_date(x):
    year = struct.unpack("!H", x[:2])[0]
    month = struct.unpack("B", x[2])[0]
    day = struct.unpack("B", x[3])[0]
    return datetime.date(year, month, day), 4

def _load_time(x):
    hour = struct.unpack("B", x[0])[0]
    minute = struct.unpack("B", x[1])[0]
    second = struct.unpack("B", x[2])[0]
    microsecond = struct.unpack("!I", '\x00' + x[3:6])[0]
    return datetime.time(hour, minute, second, microsecond), 6

def _load_datetime(x):
    return datetime.datetime.combine(
            _load_date(x)[0],
            _load_time(x[4:10])[0]), 10

def _load_timedelta(x):
    return (datetime.timedelta(
            _load_int(x)[0], _load_int(x[4:8])[0], _load_int(x[8:12])[0]), 12)

def _load_decimal(x):
    sign = _load_char(x[0])[0]
    expo = _load_short(x[1:3])[0]
    count = _load_ushort(x[3:5])[0]

    width = 5 + (count >> 1) + (count & 1)
    digit_bytes = map(ord, x[5:width])
    digits = []
    for i, b in enumerate(digit_bytes[:-1]):
        digits.append(b & 0x0f)
        digits.append(b >> 4)
    digits.append(digit_bytes[-1] & 0x0f)
    if not count & 1:
        digits.append(digit_bytes[-1] >> 4)

    return decimal.Decimal((sign, digits, expo)), width

def _load_specialnum(x):
    b = _load_uchar(x[0])[0]
    if (b & 0xf0) == MUMMY_SPECIAL_INFINITY:
        if b & 0x01:
            return decimal.Decimal("-Infinity"), 1
        return decimal.Decimal("Infinity"), 1

    if (b & 0xf0) == MUMMY_SPECIAL_NAN:
        if b & 0x01:
            return decimal.Decimal("sNaN"), 1
        return decimal.Decimal("NaN"), 1

def _load_fraction(x):
    return fractions.Fraction(*struct.unpack("!qq", x[:16])), 16


_loaders = {
    MUMMY_TYPE_NULL: _load_none,
    MUMMY_TYPE_BOOL: _load_bool,
    MUMMY_TYPE_CHAR: _load_char,
    MUMMY_TYPE_SHORT: _load_short,
    MUMMY_TYPE_INT: _load_int,
    MUMMY_TYPE_LONG: _load_long,
    MUMMY_TYPE_HUGE: _load_huge,
    MUMMY_TYPE_FLOAT: _load_double,
    MUMMY_TYPE_SHORTSTR: _load_shortstr,
    MUMMY_TYPE_MEDSTR: _load_medstr,
    MUMMY_TYPE_LONGSTR: _load_longstr,
    MUMMY_TYPE_SHORTUTF8: _load_shortutf8,
    MUMMY_TYPE_MEDUTF8: _load_medutf8,
    MUMMY_TYPE_LONGUTF8: _load_longutf8,
    MUMMY_TYPE_LONGLIST: _load_list,
    MUMMY_TYPE_LONGTUPLE: _load_tuple,
    MUMMY_TYPE_LONGSET: _load_set,
    MUMMY_TYPE_LONGHASH: _load_dict,
    MUMMY_TYPE_SHORTLIST: _load_shortlist,
    MUMMY_TYPE_MEDLIST: _load_medlist,
    MUMMY_TYPE_SHORTTUPLE: _load_shorttuple,
    MUMMY_TYPE_MEDTUPLE: _load_medtuple,
    MUMMY_TYPE_SHORTSET: _load_shortset,
    MUMMY_TYPE_MEDSET: _load_medset,
    MUMMY_TYPE_SHORTHASH: _load_shortdict,
    MUMMY_TYPE_MEDHASH: _load_meddict,
    MUMMY_TYPE_DATE: _load_date,
    MUMMY_TYPE_TIME: _load_time,
    MUMMY_TYPE_DATETIME: _load_datetime,
    MUMMY_TYPE_TIMEDELTA: _load_timedelta,
    MUMMY_TYPE_DECIMAL: _load_decimal,
    MUMMY_TYPE_SPECIALNUM: _load_specialnum,
    MUMMY_TYPE_FRACTION: _load_fraction,
}

def _loads(data):
    kind = _load_char(data)[0]
    return _loaders[kind](data[1:])

def pure_python_loads(data):
    """convert a mummy string into the python object it represents
    
    :param bytestring serialized: the serialized string to load

    :returns: the python data
    """
    if not data:
        raise ValueError("no data from which to load")
    if ord(data[0]) >> 7:
        if not lzf:
            raise RuntimeError("can't decompress without python-lzf")
        kind, ucsize, data = (
                chr(ord(data[0]) & 0x7f), _load_int(data[1:5])[0], data[5:])
        data = kind + lzf.decompress(data, ucsize + 1)

    return _loads(string(data))[0]


try:
    from _mummy import dumps, loads
    has_extension = True
except ImportError:
    dumps = pure_python_dumps
    loads = pure_python_loads
    has_extension = False



import datetime
import decimal
import itertools
import struct
import sys

try:
    import lzf
except ImportError:
    lzf = None


__all__ = ["loads", "dumps", "pure_python_loads", "pure_python_dumps",
        "has_extension"]


if sys.version_info[0] >= 3:
    from functools import reduce
    _str = unicode = str
    class string(bytes):
        def __getitem__(self, index):
            if isinstance(index, int):
                return type(self)(super().__getitem__(
                    slice(index, index + 1, None)))
            return type(self)(super().__getitem__(index))

        def __iter__(self):
            i, l = 0, len(self)
            while i < l:
                yield self[i]
                i += 1

    long = int
    map = lambda *a: list(__builtins__['map'](*a))
    iteritems = lambda d: d.items()
    iterkeys = lambda d: d.keys()
    itervalues = lambda d: d.values()
    xrange = range
    chr = lambda x: bytes([x])
    ord = lambda x: bytes.__getitem__(x, 0)
    bytify = lambda x: x.encode('ascii')
    izip = zip
else:
    iteritems = lambda d: d.iteritems()
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    bytify = lambda x: x
    bytes = string = str
    izip = itertools.izip


MAX_DEPTH = 256

TYPE_NONE = 0x0
TYPE_BOOL = 0x1
TYPE_CHAR = 0x2
TYPE_SHORT = 0x3
TYPE_INT = 0x4
TYPE_LONG = 0x5
TYPE_HUGE = 0x6
TYPE_DOUBLE = 0x7
TYPE_SHORTSTR = 0x8
TYPE_LONGSTR = 0x9
TYPE_SHORTUTF8 = 0xA
TYPE_LONGUTF8 = 0xB
TYPE_LIST = 0xC
TYPE_TUPLE = 0xD
TYPE_SET = 0xE
TYPE_DICT = 0xF

TYPE_SHORTLIST = 0x10
TYPE_SHORTTUPLE = 0x11
TYPE_SHORTSET = 0x12
TYPE_SHORTDICT = 0x13

TYPE_MEDLIST = 0x14
TYPE_MEDTUPLE = 0x15
TYPE_MEDSET = 0x16
TYPE_MEDDICT = 0x17
TYPE_MEDSTR = 0x18
TYPE_MEDUTF8 = 0x19

TYPE_DATE = 0x1A
TYPE_TIME = 0x1B
TYPE_DATETIME = 0x1C
TYPE_TIMEDELTA = 0x1D

TYPE_DECIMAL = 0x1E


TYPEMAP = {
    type(None): TYPE_NONE,
    bool: TYPE_BOOL,
    float: TYPE_DOUBLE,
}

_BIG_ENDIAN = struct.pack("!h", 1) == struct.pack("h", 1)


def _get_type_code(x):
    mapped = TYPEMAP.get(type(x))
    if mapped is not None:
        return mapped

    if type(x) is list:
        if len(x) < 256:
            return TYPE_SHORTLIST
        if len(x) < 65536:
            return TYPE_MEDLIST
        return TYPE_LIST

    if type(x) is tuple:
        if len(x) < 256:
            return TYPE_SHORTTUPLE
        if len(x) < 65536:
            return TYPE_MEDTUPLE
        return TYPE_TUPLE

    if type(x) in (set, frozenset):
        if len(x) < 256:
            return TYPE_SHORTSET
        if len(x) < 65536:
            return TYPE_MEDSET
        return TYPE_SET

    if type(x) is dict:
        if len(x) < 256:
            return TYPE_SHORTDICT
        if len(x) < 65536:
            return TYPE_MEDDICT
        return TYPE_DICT

    if type(x) is bytes:
        if len(x) < 256:
            return TYPE_SHORTSTR
        if len(x) < 65536:
            return TYPE_MEDSTR
        return TYPE_LONGSTR

    if type(x) is unicode:
        if len(x.encode('utf8')) < 256:
            return TYPE_SHORTUTF8
        if len(x.encode('utf8')) < 65536:
            return TYPE_MEDUTF8
        return TYPE_LONGUTF8

    if type(x) in (int, long):
        if -128 <= x < 128:
            return TYPE_CHAR
        if -32768 <= x < 32768:
            return TYPE_SHORT
        if -2147483648 <= x < 2147483648:
            return TYPE_INT
        if -9223372036854775808 <= x < 9223372036854775808:
            return TYPE_LONG
        return TYPE_HUGE

    if type(x) is datetime.date:
        return TYPE_DATE

    if type(x) is datetime.time:
        return TYPE_TIME

    if type(x) is datetime.datetime:
        return TYPE_DATETIME

    if type(x) is datetime.timedelta:
        return TYPE_TIMEDELTA

    if type(x) is decimal.Decimal:
        return TYPE_DECIMAL

    raise ValueError("%r cannot be serialized" % type(x))


##
## DUMPERS
##

def _dump_none(x, depth=0, default=None):
    return bytify("")

def _dump_bool(x, depth=0, default=None):
    return bytify(x and '\x01' or '\x00')

def _dump_char(x, depth=0, default=None):
    if x < 0:
        x += 256
    return chr(x)

def _dump_uchar(x, depth=0, default=None):
    return chr(x)

def _dump_short(x, depth=0, default=None):
    return struct.pack("!h", x)

def _dump_ushort(x, depth=0, default=None):
    return struct.pack("!H", x)

def _dump_int(x, depth=0, default=None):
    return struct.pack("!i", x)

def _dump_uint(x, depth=0, default=None):
    return struct.pack("!I", x)

def _dump_long(x, depth=0, default=None):
    return struct.pack("!q", x)

def _dump_huge(x, depth=0, default=None):
    data = []
    neg = x < 0
    if neg:
        x = ~x
    while x:
        data.append(x & 0xff)
        x >>= 8
    if neg:
        data = map(lambda byte: byte ^ 0xff, data)
    if not _BIG_ENDIAN:
        data = data[::-1]
    if neg and not data[0] & 0x80:
        data = [255] + data
    if not neg and data[0] & 0x80:
        data = [0] + data
    data = map(chr, data)
    return _dump_uint(len(data)) + bytify("").join(data)

def _dump_double(x, depth=0, default=None):
    return struct.pack("!d", x)

def _dump_shortstr(x, depth=0, default=None):
    return _dump_uchar(len(x)) + x

def _dump_medstr(x, depth=0, default=None):
    return _dump_ushort(len(x)) + x

def _dump_longstr(x, depth=0, default=None):
    return _dump_uint(len(x)) + x

def _dump_shortutf8(x, depth=0, default=None):
    return _dump_shortstr(x.encode('utf8'))

def _dump_medutf8(x, depth=0, default=None):
    return _dump_medstr(x.encode('utf8'))

def _dump_longutf8(x, depth=0, default=None):
    return _dump_longstr(x.encode('utf8'))

def _dump_list(x, depth=0, default=None):
    return _dump_uint(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0)
            for item in x)

_dump_set = _dump_tuple = _dump_list

def _dump_dict(x, depth=0, default=None):
    return _dump_uint(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0) for item in
            reduce(lambda a, b: a.extend(b) or a, iteritems(x), []))

def _dump_shortlist(x, depth=0, default=None):
    return _dump_uchar(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0)
            for item in x)

_dump_shorttuple = _dump_shortset = _dump_shortlist

def _dump_shortdict(x, depth=0, default=None):
    return _dump_uchar(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0) for item in
            reduce(lambda a, b: a.extend(b) or a, iteritems(x), []))

def _dump_medlist(x, depth=0, default=None):
    return _dump_ushort(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0)
            for item in x)

_dump_medtuple = _dump_medset = _dump_medlist

def _dump_meddict(x, depth=0, default=None):
    return _dump_ushort(len(x)) + bytify("").join(
            pure_python_dumps(item, default, depth + 1, compress=0) for item in
            reduce(lambda a, b: a.extend(b) or a, iteritems(x), []))

def _dump_date(x, depth=0, default=None):
    return "".join(
            (_dump_ushort(x.year), _dump_char(x.month), _dump_char(x.day)))

def _dump_time(x, depth=0, default=None):
    if x.tzinfo is not None:
        raise ValueError("can't serialize data objects with tzinfo")
    return "".join((
        _dump_char(x.hour),
        _dump_char(x.minute),
        _dump_char(x.second),
        struct.pack("!I", x.microsecond)[-3:]))

def _dump_datetime(x, depth=0, default=None):
    return _dump_date(x.date()) + _dump_time(x.timetz())

def _dump_timedelta(x, depth=0, default=None):
    return "".join((
        _dump_int(x.days), _dump_int(x.seconds), _dump_int(x.microseconds)))

def _dump_decimal(x, depth=0, default=None):
    sign, digits, expo = x.as_tuple()
    flags = 0

    flags |= expo in ("n", "N", "F")
    flags |= sign << 1
    if flags & 1:
        if expo == "F":
            flags |= 4
        else:
            flags |= (expo == "N") << 3

        return _dump_char(flags)

    digitpairs = []
    for i, dig in enumerate(digits):
        if not 0 <= dig <= 9:
            raise ValueError("invalid digit")

        if not (i & 1):
            digitpairs.append(0)
            dig <<= 4

        digitpairs[-1] |= dig

    return (struct.pack("!BhH", flags, expo, len(digits)) +
            "".join(map(chr, digitpairs)))


_dumpers = {
    TYPE_NONE: _dump_none,
    TYPE_BOOL: _dump_bool,
    TYPE_CHAR: _dump_char,
    TYPE_SHORT: _dump_short,
    TYPE_INT: _dump_int,
    TYPE_LONG: _dump_long,
    TYPE_HUGE: _dump_huge,
    TYPE_DOUBLE: _dump_double,
    TYPE_SHORTSTR: _dump_shortstr,
    TYPE_MEDSTR: _dump_medstr,
    TYPE_LONGSTR: _dump_longstr,
    TYPE_SHORTUTF8: _dump_shortutf8,
    TYPE_MEDUTF8: _dump_medutf8,
    TYPE_LONGUTF8: _dump_longutf8,
    TYPE_LIST: _dump_list,
    TYPE_TUPLE: _dump_tuple,
    TYPE_SET: _dump_set,
    TYPE_DICT: _dump_dict,
    TYPE_SHORTLIST: _dump_shortlist,
    TYPE_MEDLIST: _dump_medlist,
    TYPE_SHORTTUPLE: _dump_shorttuple,
    TYPE_MEDTUPLE: _dump_medtuple,
    TYPE_SHORTSET: _dump_shortset,
    TYPE_MEDSET: _dump_medset,
    TYPE_SHORTDICT: _dump_shortdict,
    TYPE_MEDDICT: _dump_meddict,
    TYPE_DATE: _dump_date,
    TYPE_TIME: _dump_time,
    TYPE_DATETIME: _dump_datetime,
    TYPE_TIMEDELTA: _dump_timedelta,
    TYPE_DECIMAL: _dump_decimal,
}

def pure_python_dumps(item, default=None, depth=0, compress=True):
    "serialize a native python object into a mummy string"
    if default and not hasattr(default, "__call__"):
        raise TypeError("default must be callable or None")
    if depth >= MAX_DEPTH:
        raise ValueError("max depth exceeded")
    try:
        kind = _get_type_code(item)
    except ValueError:
        if default is None:
            raise TypeError("unserializable type")
        item = default(item)
        kind = _get_type_code(item)
    data = _dumpers[kind](item, depth, default)
    datalen = len(data)
    if compress and lzf and datalen > 5:
        compressed = lzf.compress(data, datalen - 5)
        if compressed:
            data = struct.pack("!i", datalen) + compressed
            kind = kind | 0x80
    kind = _dump_char(kind)

    return kind + data


##
## LOADERS
##

def _load_none(x):
    return None, 0

def _load_bool(x):
    return bool(ord(x[0])), 1

def _load_char(x):
    num = ord(x[0])
    if num >= 128:
        num -= 256
    return num, 1

def _load_uchar(x):
    return ord(x[0]), 1

def _load_short(x):
    return struct.unpack("!h", x[:2])[0], 2

def _load_ushort(x):
    return struct.unpack("!H", x[:2])[0], 2

def _load_int(x):
    return struct.unpack("!i", x[:4])[0], 4

def _load_uint(x):
    return struct.unpack("!I", x[:4])[0], 4

def _load_long(x):
    return struct.unpack("!q", x[:8])[0], 8

def _load_huge(x):
    width = _load_uint(x)[0] + 4
    num = 0
    x = x[4:]
    neg = ord(x[0]) & 0x80
    data = map(ord, x[:width])
    if neg:
        data = map(lambda byte: byte ^ 0xff, data)
    for c in data:
        num = (num << 8) | c
    if neg:
        return -num - 1, width
    return num, width

def _load_double(x):
    return struct.unpack("!d", x[:8])[0], 8

def _load_shortstr(x):
    width = _load_uchar(x)[0] + 1
    return x[1:width], width

def _load_medstr(x):
    width = _load_ushort(x)[0] + 2
    return x[2:width], width

def _load_longstr(x):
    width = _load_uint(x)[0] + 4
    return x[4:width], width

def _load_shortutf8(x):
    str, width = _load_shortstr(x)
    return str.decode('utf8'), width

def _load_medutf8(x):
    str, width = _load_medstr(x)
    return str.decode('utf8'), width

def _load_longutf8(x):
    str, width = _load_longstr(x)
    return str.decode('utf8'), width

def _load_list(x):
    length, width = _load_uint(x)
    result = []
    for i in xrange(length):
        item, item_width = _loads(x[width:])
        result.append(item)
        width += item_width + 1
    return result, width

def _load_set(x):
    lst, width = _load_list(x)
    return set(lst), width

def _load_tuple(x):
    lst, width = _load_list(x)
    return tuple(lst), width

def _load_dict(x):
    length, width = _load_uint(x)
    result = {}
    for i in xrange(length):
        key, keywidth = _loads(x[width:])
        width += keywidth + 1
        value, valuewidth = _loads(x[width:])
        width += valuewidth + 1
        result[key] = value
    return result, width

def _load_shortlist(x):
    length, width = _load_uchar(x)
    result = []
    for i in xrange(length):
        item, item_width = _loads(x[width:])
        result.append(item)
        width += item_width + 1
    return result, width

def _load_shorttuple(x):
    lst, width = _load_shortlist(x)
    return tuple(lst), width

def _load_shortset(x):
    lst, width = _load_shortlist(x)
    return set(lst), width

def _load_shortdict(x):
    length, width = _load_uchar(x)
    result = {}
    for i in xrange(length):
        key, keywidth = _loads(x[width:])
        width += keywidth + 1
        value, valuewidth = _loads(x[width:])
        width += valuewidth + 1
        result[key] = value
    return result, width

def _load_medlist(x):
    length, width = _load_ushort(x)
    result = []
    for i in xrange(length):
        item, item_width = _loads(x[width:])
        result.append(item)
        width += item_width + 1
    return result, width

def _load_medtuple(x):
    lst, width = _load_medlist(x)
    return tuple(lst), width

def _load_medset(x):
    lst, width = _load_medlist(x)
    return set(lst), width

def _load_meddict(x):
    length, width = _load_ushort(x)
    result = {}
    for i in xrange(length):
        key, keywidth = _loads(x[width:])
        width += keywidth + 1
        value, valuewidth = _loads(x[width:])
        width += valuewidth + 1
        result[key] = value
    return result, width

def _load_date(x):
    year = struct.unpack("!H", x[:2])[0]
    month = struct.unpack("B", x[2])[0]
    day = struct.unpack("B", x[3])[0]
    return datetime.date(year, month, day), 4

def _load_time(x):
    hour = struct.unpack("B", x[0])[0]
    minute = struct.unpack("B", x[1])[0]
    second = struct.unpack("B", x[2])[0]
    microsecond = struct.unpack("!I", '\x00' + x[3:6])[0]
    return datetime.time(hour, minute, second, microsecond), 6

def _load_datetime(x):
    return datetime.datetime.combine(
            _load_date(x)[0],
            _load_time(x[4:10])[0]), 10

def _load_timedelta(x):
    return (datetime.timedelta(
            _load_int(x)[0], _load_int(x[4:8])[0], _load_int(x[8:12])[0]), 12)

def _load_decimal(x):
    flags = struct.unpack("B", x[0])[0]

    if flags & 1:
        width = 1
        if flags & 4:
            # (+ or -) Infinity
            triple = ((flags & 2) >> 1, (0,), "F")
        else:
            # [s]NaN
            triple = (0, (), flags & 8 and "N" or "n")
    else:
        sign = (flags & 2) >> 1
        exponent, length = struct.unpack("!hH", x[1:5])
        width = 5 + (length // 2) + (length & 1)

        digitbytes = map(ord, x[5:width])
        digits = []
        for i, b in enumerate(digitbytes):
            digits.append((b & 0xf0) >> 4)
            digits.append(b & 0xf)

        if not digits[-1]:
            digits.pop()

        triple = (sign, digits, exponent)

    return decimal.Decimal(triple), width


_loaders = {
    TYPE_NONE: _load_none,
    TYPE_BOOL: _load_bool,
    TYPE_CHAR: _load_char,
    TYPE_SHORT: _load_short,
    TYPE_INT: _load_int,
    TYPE_LONG: _load_long,
    TYPE_HUGE: _load_huge,
    TYPE_DOUBLE: _load_double,
    TYPE_SHORTSTR: _load_shortstr,
    TYPE_MEDSTR: _load_medstr,
    TYPE_LONGSTR: _load_longstr,
    TYPE_SHORTUTF8: _load_shortutf8,
    TYPE_MEDUTF8: _load_medutf8,
    TYPE_LONGUTF8: _load_longutf8,
    TYPE_LIST: _load_list,
    TYPE_TUPLE: _load_tuple,
    TYPE_SET: _load_set,
    TYPE_DICT: _load_dict,
    TYPE_SHORTLIST: _load_shortlist,
    TYPE_MEDLIST: _load_medlist,
    TYPE_SHORTTUPLE: _load_shorttuple,
    TYPE_MEDTUPLE: _load_medtuple,
    TYPE_SHORTSET: _load_shortset,
    TYPE_MEDSET: _load_medset,
    TYPE_SHORTDICT: _load_shortdict,
    TYPE_MEDDICT: _load_meddict,
    TYPE_DATE: _load_date,
    TYPE_TIME: _load_time,
    TYPE_DATETIME: _load_datetime,
    TYPE_TIMEDELTA: _load_timedelta,
    TYPE_DECIMAL: _load_decimal,
}

def _loads(data):
    kind = _load_char(data)[0]
    return _loaders[kind](data[1:])

def pure_python_loads(data):
    "convert a mummy string into the python object it represents"
    if not data:
        raise ValueError("no data from which to load")
    if ord(data[0]) >> 7:
        if not lzf:
            raise RuntimeError("can't decompress without python-lzf")
        kind, ucsize, data = (
                chr(ord(data[0]) & 0x7f), _load_int(data[1:5])[0], data[5:])
        data = kind + lzf.decompress(data, ucsize + 1)

    return _loads(string(data))[0]


try:
    from _oldmummy import dumps, loads
    has_extension = True
except ImportError:
    dumps = pure_python_dumps
    loads = pure_python_loads
    has_extension = False



"""
mummy is a data serialization format and accompanying (de)serializer.

the format is designed to be first fast, and second compact, willing to
sacrifice human-readability in the name of those two values.

the main implementation is in C, but there is a pure-python version it falls
back to if the extension is unavailable. the module-global `has_extension` is a
boolean indicating whether the C extension is in use.
"""

from __future__ import absolute_import

from .serialization import \
        loads, dumps, pure_python_loads, pure_python_dumps, has_extension
from .schemas import Message, OPTIONAL, UNION, ANY


__all__ = ["loads", "dumps", "pure_python_loads", "pure_python_dumps",
        "has_extension", "Message", "OPTIONAL", "UNION", "ANY"]



"""
SCHEMAS:

simple schemas:
    - just a basic atomic python type (the type object itself)
        - bool
        - int (also validates longs in python 2.x)
        - float
        - str
    - asserts that the validated object is an instance

simple instance schemas:
    - an instance of any of the atomic python types (bool, int, float, str)
    - asserts that the validated object is equal

tuple schemas:
    - assert that the validated object is a tuple
    - assets that the validated tuple is the same length (with a caveat related
      to OPTIONALs, see below)
    - pairs up the items of the validated tuple with the sub-schemas in the
      schema tuple and asserts that they all match
    - sub-schemas may be OPTIONALs, in which case they don't have to be present
      at all in the validated tuple
        - all OPTIONALs must come after all non-OPTIONALs so that it is
          possible to match everything up
        - the n objects in the validated tuple after all non-OPTIONALs will be
          matched up with the *first* n OPTIONALs in the schema tuple in order
          (so for example the second OPTIONAL may be be present unless the
          first one is)

list schemas:
    - assert that the validated object is itself a list
    - only homogeneous lists may be validated, so the schema list must be of
      length 0 or 1 (although "homogeneous" is in terms of the sub-schemata,
      which may actually be UNIONs)
    - if the schema list is empty it will only match an empty list
    - the item in the list schema is a schema which is used to match all the
      contents of the validated list
    - if the item in the schema list is an OPTIONAL, the list is allowed to be
      empty

dict schemas:
    - assert that the validated object is a dict
    - values are schemas used to validate the validated dict's values
    - keys are used to match up the sub-schemas to their values
    - keys may be instances of simple types (bool, int, float, str)
        - asserts that the exact key is present, and matches the corresponding
          value to the sub-schema
    - keys may be simple type objects (the types, not instances)
        - matches any key of that type that is not matched by a simple type
          instance or tuple
        - by extension, allows any keys of this type in the validated dict (as
          long as their values match the paired sub-schema)
    - if a value is an OPTIONAL, specifies that the corresponding key doesn't
      necessarily have to match anything in the validated object

UNION schemas:
    - instances of UNION, the constructor of which accepts any number of
      sub-schemas as arguments
    - attempts matching the message against each of the sub-schemas, validates
      if any of them match

OPTIONAL schemas:
    - only usable inside compound schema types (tuple, list, dict)
    - specifies that a particular slot in the compound schema is optional

ANY schemas:
    - ANY is a singleton that can be used as a schema simply to consider
      anything as valid

RULE schemas:
    - the RULE constructor accepts a function as an argument; the function
      should take one argument and return a boolean. the schema requires that
      the function, when called on the matched message, returns True


A schema definition can be any of the above schema types and any allowed
sub-schemas.

>>> address_book_schema = [{
...     'first_name': str,
...     'last_name': str,
...     'is_male': bool,
...     'birthday': datetime.date,
...     'address': {
...         'street_name': str,
...         'street_number': int,
...         OPTIONAL('sub_number'): str, # apt # or similar (eg the B in 345B)
...         'zip_code': int,
...         'city': str,
...         OPTIONAL('state'): str, # optional for countries without states
...         'country': str,
...     },
...     'hobbies': [OPTIONAL(str)],
...     'properties': ANY,
... }]


The API for schemas is just 5 names: OPTIONAL, UNION, ANY, RULE, and Message.
The first 4 were explained above. Message is a base class you can use to create
message classes; just give them a SCHEMA attribute of a valid schema and they
can be used to validate, shorten, and serialize their instances.

>>> import mummy
>>> class AddressBookMessage(mummy.Message):
...     SCHEMA = address_book_schema
...

abm = AddressBookMessage([{
    'first_name': 'Travis',
    'last_name': 'Parker',
    'is_male': True,
    'birthday': datetime.date(1984, 1, 6),
    'address': {
        'street_number': 11,
        'zip_code': 12345,
        'street_name': 'None',
        'city': 'Of',
        'state': 'Your',
        'country': 'Business',
    },
    'hobbies': [],
    'properties': None,
}])

>>> abm.validate()

>>> abm.message
[{'first_name': 'Travis', 'last_name': 'Parker', 'is_male': True, 'hobbies': [], 'birthday': datetime.date(1984, 1, 6), 'address': {'city': 'Of', 'street_number': 11, 'country': 'Business', 'street_name': 'None', 'state': 'Your', 'zip_code': 12345}, 'properties': None}]


Instances of the message class are able to shorten the serialized data by
removing information that is already present in the schema itself, and the
message class itself is able to undo that transformation.

>>> abm.transform()
[[['Of', 'Business', 'None', 11, 12345, 'Your', None], datetime.date(1984, 1, 6), 'Travis', [], True, 'Parker', None]]
>>> AddressBookMessage.untransform(abm.transform()) == abm.message
True


The result is that messages serialized through a Message class are much shorter
than if the same data were serialized via basic mummy.dumps, but as long as the
message receiver has the schema as well, no information is lost.

>>> len(abm.dumps())
64
>>> len(mummy.dumps(abm.message))
195
>>> AddressBookMessage.loads(abm.dumps()).message == abm.message
True

"""

from __future__ import absolute_import

import datetime
import decimal
import itertools
import sys

from .serialization import loads, dumps


__all__ = ["Message", "OPTIONAL", "UNION", "ANY"]


# python 2/3 compat stuff
if sys.version_info[0] >= 3:
    izip = zip
    imap = map
    ifilter = filter
    def iteritems(d):
        return d.items()
    def iterkeys(d):
        return d.keys()
    def itervalues(d):
        return d.values()
    long = int
else:
    izip = itertools.izip
    imap = itertools.imap
    ifilter = itertools.ifilter
    def iteritems(d):
        return d.iteritems()
    def iterkeys(d):
        return d.iterkeys()
    def itervalues(d):
        return d.itervalues()


_primitives = (
        bool, int, float, str,
        datetime.date, datetime.time, datetime.datetime, datetime.timedelta,
        decimal.Decimal)
_type_validations = {
    bool: bool,
    int: (int, long),
    float: float,
    str: str,
    datetime.date: datetime.date,
    datetime.time: datetime.time,
    datetime.datetime: datetime.datetime,
    datetime.timedelta: datetime.timedelta,
    decimal.Decimal: decimal.Decimal,
}

class OPTIONAL(object):
    """specifies that a piece of a schema is optional in some specific contexts

    - as a dictionary key, allows that key to be left out
    - as a member of a tuple schema, allows the message tuple to be left out
    - as the only member of a list schema, allows the message list to be empty
    """
    def __init__(self, schema):
        self.schema = schema

    def __repr__(self):
        return "OPTIONAL(%r)" % (self.schema,)

def _optional(x):
    return isinstance(x, OPTIONAL)

def _required(x):
    return not isinstance(x, OPTIONAL)

class UNION(object):
    "specify that a message may match any of the sub-schemas"
    def __init__(self, *options):
        self.options = options

    def __repr__(self):
        return "UNION(%r)" % (list(self.options),)

class ANY(object):
    "validates any object successfully"
    def __call__(self):
        return self

    def __repr__(self):
        return "<ANY>"
ANY = ANY()

class RULE(object):
    "require that a message passes a given boolean predicate"
    def __init__(self, pred):
        self.pred = pred

    def __repr__(self):
        return "<RULE (%r)>" % self.pred


##
## Validation
##

def _validate_simple(schema, message):
    if isinstance(message, _type_validations[schema]):
        return True, None
    return False, (message, schema)

def _validate_simple_instance(schema, message):
    if schema == message:
        return True, None
    return False, (message, schema)

def _validate_tuple(schema, message):
    if not isinstance(message, tuple):
        return False, (message, schema)

    if len(message) > len(schema):
        return False, (message, schema)

    required = izip(
            itertools.takewhile(_required, schema),
            message)

    for i, (sub_schema, sub_message) in enumerate(required):
        matched, info = _validate(sub_schema, sub_message)
        if not matched:
            return False, info

    optional = itertools.islice(izip(schema, message), i + 1, None)

    for sub_schema, sub_message in optional:
        matched, info = _validate(sub_schema, sub_message)
        if not matched:
            return False, info

    return True, None

def _validate_list(schema, message):
    if not isinstance(message, list):
        return False, (message, schema)

    if not schema:
        if message:
            return False, (message, schema)
        return True, None

    sub_schema = schema[0]
    if isinstance(sub_schema, OPTIONAL):
        if not message:
            return True, None
        sub_schema = sub_schema.schema
    elif not message:
        return False, (message, schema)

    for sub_message in message:
        matched, info = _validate(sub_schema, sub_message)
        if not matched:
            return False, info

    return True, None

def _validate_dict(schema, message):
    if not isinstance(message, dict):
        return False, (message, schema)

    required_keys = set(k for k in schema if not isinstance(k, OPTIONAL))
    required_wildcards = required_keys.intersection(_primitives)
    required_keys.difference_update(_primitives)

    msg_keys = set(message)

    # missing required keys
    if required_keys - msg_keys:
        return False, (message, schema)

    # missing required wildcards
    if required_wildcards - set(imap(type, message)):
        return False, (message, schema)

    wildcards = set(_primitives)
    wildcards.intersection_update(schema)

    # extra non-allowed keys
    msg_keys.difference_update(schema)
    msg_keys.difference_update(
            k.schema for k in schema if isinstance(k, OPTIONAL))
    if set(imap(type, msg_keys)) - wildcards:
        return False, (message, schema)

    schema = schema.copy()
    schema.update(
        dict((k.schema, v) for k, v in iteritems(schema)
            if isinstance(k, OPTIONAL)))

    # now validate sub_schemas/sub_messages
    for key, sub_message in iteritems(message):
        if key in schema:
            sub_schema = schema[key]
        else:
            sub_schema = schema[type(key)]

        matched, info = _validate(sub_schema, sub_message)
        if not matched:
            return False, info

    return True, None

def _validate_union(schema, message):
    for sub_schema in schema.options:
        matched, info = _validate(sub_schema, message)
        if matched:
            return True, None
    return False, (message, schema)

def _validate_any(schema, message):
    return True, None

def _validate_rule(schema, message):
    if schema.pred(message):
        return True, None
    return False, (message, schema)

_validators = {
    bool: _validate_simple_instance,
    int: _validate_simple_instance,
    long: _validate_simple_instance,
    float: _validate_simple_instance,
    str: _validate_simple_instance,
    type: _validate_simple,
    tuple: _validate_tuple,
    list: _validate_list,
    dict: _validate_dict,
    UNION: _validate_union,
    type(ANY): _validate_any,
    RULE: _validate_rule,
}

def _validate(schema, message):
    return _validators[type(schema)](schema, message)


##
## Validation of the Schema itself
##

def _validate_simple_instance_schema(schema):
    return True, None

def _validate_simple_schema(schema):
    if schema in _type_validations:
        return True, None
    return False, schema

def _validate_tuple_schema(schema):
    for sub_schema in schema:
        if isinstance(sub_schema, OPTIONAL):
            sub_schema = sub_schema.schema
        valid, info = _validate_schema(sub_schema)
        if not valid:
            return False, info
    return True, None

def _validate_list_schema(schema):
    if not schema:
        return True, None

    if len(schema) != 1:
        return False, schema

    sub_schema = schema[0]
    if isinstance(sub_schema, OPTIONAL):
        sub_schema = sub_schema.schema

    return _validate_schema(sub_schema)

def _validate_dict_schema(schema):
    for key in iterkeys(schema):
        if isinstance(key, OPTIONAL):
            key = key.schema

        if isinstance(key, _primitives + (long,)):
            continue

        if key in _primitives:
            continue

        return False, schema

    for sub_schema in itervalues(schema):
        valid, info = _validate_schema(sub_schema)
        if not valid:
            return False, info

    return True, None

def _validate_union_schema(schema):
    for sub_schema in schema.options:
        valid, info = _validate_schema(sub_schema)
        if not valid:
            return False, info
    return True, None

def _validate_any_schema(schema):
    return True, None

def _validate_rule_schema(schema):
    if callable(getattr(schema, "pred", None)):
        return True, None
    return False, schema

_schema_validators = {
    bool: _validate_simple_instance_schema,
    int: _validate_simple_instance_schema,
    long: _validate_simple_instance_schema,
    float: _validate_simple_instance_schema,
    str: _validate_simple_instance_schema,
    type: _validate_simple_schema,
    tuple: _validate_tuple_schema,
    list: _validate_list_schema,
    dict: _validate_dict_schema,
    UNION: _validate_union_schema,
    type(ANY): _validate_any_schema,
    RULE: _validate_rule_schema,
}

def _validate_schema(schema):
    schema_type = type(schema)
    if schema_type not in _schema_validators:
        return False, schema
    return _schema_validators[schema_type](schema)


##
## [un]transforming messages based on a schema
##

def _group_schema_keys(schema):
    key_set = set(schema)

    key_set.difference_update(_type_validations)

    optional = [k for k in schema if isinstance(k, OPTIONAL)]
    key_set.difference_update(optional)
    optional = [k.schema for k in optional]
    optional.sort()

    required = list(key_set)
    required.sort()

    return required, optional

def _transform(schema, message):
    if isinstance(schema, list):
        return [_transform(s, m) for s, m in izip(schema, message)]

    if isinstance(schema, tuple):
        return tuple(_transform(s, m) for s, m in izip(schema, message))

    if not isinstance(schema, dict):
        return message

    required, optional = _group_schema_keys(schema)

    msgkeys = set(message)

    schema = schema.copy()
    schema.update(
            dict((k.schema, v) for k, v in iteritems(schema)
                if isinstance(k, OPTIONAL)))

    result = []
    for key in required:
        result.append(_transform(schema[key], message[key]))
    for key in optional:
        if key in message:
            result.append(_transform(schema[key], message[key]))
        else:
            result.append(None)
    msgkeys.difference_update(required)
    msgkeys.difference_update(optional)

    for key in msgkeys:
        result.append(key)
        result.append(_transform(schema[type(message[key])], message[key]))

    return result

def _untransform(schema, message):
    if isinstance(schema, list):
        return [_untransform(s, m) for s, m in izip(schema, message)]

    if isinstance(schema, tuple):
        return tuple(_untransform(s, m) for s, m in izip(schema, message))

    if not isinstance(schema, dict):
        return message

    required, optional = _group_schema_keys(schema)
    result = {}

    for key, value in izip(required, message):
        result[key] = _untransform(schema[key], value)

    schema = schema.copy()
    schema.update(
            dict((k.schema, v) for k, v in iteritems(schema)
                if isinstance(k, OPTIONAL)))

    for key, value in izip(optional, itertools.islice(
            message, len(required), None)):
        if value is not None:
            result[key] = _untransform(schema[key], value)

    is_key = True
    for item in itertools.islice(
            message, len(required) + len(optional), None):
        if is_key:
            key = item
        else:
            result[key] = _untransform(schema[type(item)], item)
        is_key = not is_key

    return result


##
## the schema metaclass
##

class _Invalid(Exception):
    pass

class InvalidSchema(_Invalid):
    pass


class _validated_schema(type):
    def __init__(cls, *args, **kwargs):
        super(_validated_schema, cls).__init__(*args, **kwargs)
        if hasattr(cls, "SCHEMA"):
            valid, info = _validate_schema(cls.SCHEMA)
            if not valid:
                raise InvalidSchema(info)

        cls.InvalidMessage = type('InvalidMessage', (_Invalid,), {})

class Message(object):
    __metaclass__ = _validated_schema

    def __init__(self, message):
        self.message = message
        self._validation = None
        self._transformation = None

    def validate(self):
        if self._validation is None:
            self._validation = _validate(self.SCHEMA, self.message)
        if not self._validation[0]:
            raise self.InvalidMessage(self._validation[1])

    def transform(self):
        self.validate()
        if self._transformation is None:
            self._transformation = _transform(self.SCHEMA, self.message)
        return self._transformation

    def dumps(self):
        self.validate()
        return dumps(self.transform())

    @classmethod
    def untransform(cls, message):
        return _untransform(cls.SCHEMA, message)

    @classmethod
    def loads(cls, message):
        return cls(cls.untransform(loads(message)))



import mummy
from StringIO import StringIO

from django.core.serializers.python import Serializer as PythonSerializer
from django.core.serializers.python import Deserializer as PythonDeserializer
from django.utils import datetime_safe
try:
    import decimal
except ImportError:
    from django.utils import _decimal as decimal # Python 2.3 fallback


class Serializer(PythonSerializer):
    internal_use_only = False

    DATE_FORMAT = "%Y-%m-%d"
    TIME_FORMAT = "%H:%M:%S"

    def end_serialization(self):
        self.options.pop('stream', None)
        self.options.pop('fields', None)
        self.stream.write(mummy.dumps(self.objects))

    def getvalue(self):
        if callable(getattr(self.stream, 'getvalue', None)):
            return self.stream.getvalue()


def Deserializer(stream_or_string, **options):
    if isinstance(stream_or_string, basestring):
        stream = StringIO(stream_or_string)
    else:
        stream = stream_or_string
    for obj in PythonDeserializer(mummy.loads(stream.read())):
        yield obj
